/*
  RabidVWM LiteStep Module - version 1.0
	Module Startup and Shutdown

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// RabidVWM.cpp : Defines the entry point for the DLL application.
//

// TAB SPACING set to 2

#include "RabidVWM.h"
#include "multimonitor.h"
#include <shellapi.h>

#include "droptarget.h"

#define VWM_DESKNO 0x11110000

#define VWM_LSBOX_CFG "VWMLoadInBox"
#define VWM_MULTIMON_CFG "VWMSwitchMonitors"

s_GlobalData GlobalData;

#ifndef NDEBUG
bool bInLocalInit = false;
DWORD TraceStartTime;
#endif

#ifndef LSTOOLS_REV
struct rcLinkItem {
	const char *name;		// the name of the function/library
	int **ptr;					// the location to store its address or NULL for lib name
};
#else
typedef lsLinkItem rcLinkItem;
#endif

HWND WINAPI GetAncestor95(HWND hWnd, UINT rootOwner); // in windows.cpp
int localParseCoordinate(LPSTR strVal, int defaultVal, int maxVal);

void (WINAPI *dSwitchToThisWindow)(HWND, int) = NULL;
BOOL (WINAPI *dIsHungAppWindow) (HWND) = NULL;
BOOL (WINAPI *dIsHungThread) (DWORD) = NULL;
BOOL  (WINAPI *dAlphaBlend)( IN HDC, IN int, IN int, IN int, IN int, IN HDC, IN int, IN int, IN int, IN int, IN BLENDFUNCTION) = NULL;
HWND (WINAPI *dGetAncestor)(HWND, UINT) = &GetAncestor95; // Not in win95, need to emulate
BOOL (WINAPI *dEnumDisplayMonitors)(IN HDC,IN LPCRECT,IN MONITORENUMPROC,IN LPARAM) = NULL;
BOOL (WINAPI *dGetMonitorInfo)( IN HMONITOR, OUT LPMONITORINFO ) = NULL;
HMONITOR (WINAPI *dMonitorFromPoint)(IN POINT,IN DWORD ) = NULL;
BOOL (WINAPI *dUpdateLayeredWindow)(HWND,HDC,POINT*,SIZE*,HDC,POINT*,COLORREF,BLENDFUNCTION*,DWORD) = NULL;
LSAPI int (*dParseCoordinate)(char strVal[], int defaultVal, int maxVal) = &localParseCoordinate;

rcLinkItem DynFuncs[] =
{
	// force window switch in nt
	{ "SwitchToThisWindow", reinterpret_cast<int**>(&dSwitchToThisWindow) },
	// check if window is frozen in nt
	{ "IsHungAppWindow", reinterpret_cast<int**>(&dIsHungAppWindow) },
	// check if thread is frozen in 9x
	{ "IsHungThread", reinterpret_cast<int**>(&dIsHungThread) },
	// Not in win95, need to emulate
	{ "GetAncestor", reinterpret_cast<int**>(&dGetAncestor) },
	// multimonitor stuff
	{ "EnumDisplayMonitors", reinterpret_cast<int**>(&dEnumDisplayMonitors) },
#ifndef _UNICODE
	{ "GetMonitorInfoA", reinterpret_cast<int**>(&dGetMonitorInfo) },
#endif
	{ "MonitorFromPoint", reinterpret_cast<int**>(&dMonitorFromPoint) },
	// win2k and newer
	{ "UpdateLayeredWindow", reinterpret_cast<int**>(&dUpdateLayeredWindow) },
//	{ "", reinterpret_cast<int**>(&d) },

	// alpha blend, not defined in win95, so has to be dynamically linked properly
	{ "MSIMG32.DLL", NULL },
	{ "AlphaBlend", reinterpret_cast<int**>(&dAlphaBlend) },

#ifndef USE_PARSECOORDINATE
	{ "LSAPI.DLL", NULL },
	{ "ParseCoordinate", reinterpret_cast<int**>(&dParseCoordinate) },
#endif

	{ NULL, NULL }
};

#ifndef LSTOOLS_REV
void rLinkFunctions(rcLinkItem *list)
{
	HINSTANCE hDLL;
  int *ptr;

	ASSERTV(list!=NULL);

	// most common module to import from, so don't require its name
	hDLL = GetModuleHandle("USER32.DLL");

	while (list->name!=NULL) {
		if (list->ptr==NULL) {
			// it's a library name

			// load it (always load so I know which need to be unloaded, all of them)
			hDLL = LoadLibrary(list->name);

#if !defined(NDEBUG) && defined(TRACE_FILE)
      {
        char buf[256];
        sprintf(buf,"DLL: %s = %08X",list->name,hDLL);
        TRACE(TRACE_DUMP,buf);
      }
#endif
		} else {
			// it's a function name

			ptr = reinterpret_cast<int *>(GetProcAddress(hDLL, list->name));
      if (ptr) *list->ptr = ptr;

#if !defined(NDEBUG) && defined(TRACE_FILE)
      {
        char buf[256];
        sprintf(buf,"Func: %s = %08X",list->name,*list->ptr);
        TRACE(TRACE_DUMP,buf);
      }
#endif
		}

		// next item
		++list;
	}
}

void rUnlinkFunctions(rcLinkItem *list)
{
	ASSERTV(list!=NULL);

	while (list->name!=NULL) {
		if (list->ptr==NULL) {
			// it's a library name

			// unload it
			FreeLibrary(GetModuleHandle(list->name));

		} else {
			// it's a function name
			// nothing to do
		}

		// next item
		++list;
	}
}
#else
inline void rLinkFunctions(rcLinkItem *list)
{ LSLinkFunctions(list); }
inline void rUnlinkFunctions(rcLinkItem *list)
{ LSUnlinkFunctions(list); }
#endif

/*                                                  --------- initModuleEx
	Entry point from LiteStep.

	Input:
		ParentWnd = litestep hwnd, unless called via initWharfModule
								should be used as parent for module window
		dllInst = my instance
		szPath = path to litestep.exe (including file name?)
	Output:
		returns 0 on success, nonzero on failure
*/
int initModuleEx (HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	int ret;

	// Do things here that need to be done ONCE when the module is first loaded
	// things that should happen if the module is refreshed should be in LocalInit

#if !defined(NDEBUG) && defined(TRACE_FILE)
	TraceStartTime = 0;
	TRACE(TRACE_STARTUP,"Hello.");
	TraceStartTime = GetTickCount();
	{
		char buf[256];
		sprintf(buf,"initModuleEx loaded at %08X",&initModuleEx);
		TRACE(TRACE_STATUS,buf);
	}
#endif
#ifdef VWM_STATS
	VWMStats.InitStart = GetTickCount();
#endif
	TRACE(TRACE_STARTUP,szAppName);
	TRACE(TRACE_STARTUP,szAppVer);
	TRACE(TRACE_STARTUP,szAppAuth);
	TRACE(TRACE_STARTUP,"");

	/* make sure it's all zeroed */ 
	ZeroMemory(&GlobalData,sizeof(s_GlobalData));

	/* save instance handle      */ 
	GlobalData.hInstance = dllInst;

	rLinkFunctions(DynFuncs);

	/* alloc known windows array */ 
	GlobalData.nKnownWindows = 0;
	AllocateWindows(MAX_WINDOWS);

	GlobalData.hLiteStepWnd = GetLitestepWnd();
	CTRACE(TRACE_WARNING,GlobalData.hLiteStepWnd==NULL,"Failed to get Litestep window.");

	TRACE(TRACE_STATUS,"Getting parent window.");
	if (ParentWnd == GlobalData.hLiteStepWnd) {
		/* using the litestep window as a parent will get me dropped behind  */ 
		/* the desktop if i'm in "always on bottom" mode, so use the desktop */ 
		TRACE(TRACE_STATUS,"Using DesktopBackgroundClass.");
		GlobalData.hParentWnd = FindWindow("DesktopBackgroundClass", NULL);
	} else {
		TRACE(TRACE_STATUS,"Using parent passed to initModuleEx.");
		GlobalData.hParentWnd = ParentWnd;
		if (!GlobalData.hParentWnd) {
			TRACE(TRACE_STATUS,"Never mind, using DesktopBackgroundClass.");
			GlobalData.hParentWnd = FindWindow("DesktopBackgroundClass", NULL);
		}
	}
	if (!GlobalData.hParentWnd) {
		TRACE(TRACE_STATUS,"Bah, using GetDesktopWindow.");
		GlobalData.hParentWnd = GetDesktopWindow();
	}

	GetWindowRect(GlobalData.hParentWnd, &GlobalData.ParentRect);

	GlobalData.bStayChildWindow = (GlobalData.hParentWnd == ParentWnd);

	/* register window class */ 
	TRACE(TRACE_STATUS,"Registering window class.");
	{
		WNDCLASS wc;

		memset(&wc,0,sizeof(wc));
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.lpfnWndProc = &WndProc;
		wc.hInstance = GlobalData.hInstance;
		wc.lpszClassName = szAppName;

		if (!RegisterClass(&wc)) 
		{
			LASTERROR(0/*RegisterClass*/);
			MessageBox(GlobalData.hParentWnd,
				"Error registering window class",
				szAppName,MB_OK|MB_ICONERROR);
			//PostQuitMessage(0); // to what window, exactly?
			return 1;
		}
	}

  /* this should allow drag&drop to work while threaded */ 
  CoInitialize(NULL);

	GlobalData.CurrentDesk = DESK_NONE;
	SendMessage(GlobalData.hLiteStepWnd, LM_RESTOREDATA,
		(WPARAM) (sizeof(int) | VWM_DESKNO), /* length and ident */ 
		(LPARAM) &GlobalData.CurrentDesk);   /* data             */ 
	if (GlobalData.CurrentDesk<0) {
		GlobalData.bGuessDeskNo = true;
		GlobalData.CurrentDesk = 0;
	}

#if defined(RABIDVWM_MULTIMON) && defined(NO_RCMM)
	LSInitMultiMon();
#endif

#ifdef VWMDROPTARGET
	VWMDropTarget::RegisterFormats();
#endif

	GlobalData.bLoadInBox = GetRCBool(VWM_LSBOX_CFG,TRUE)!=FALSE;
	/* if we're loading in lsbox, don't read the config yet */ 
	if (GlobalData.bLoadInBox) {
		TRACE(TRACE_STATUS,"Waiting for !VWMLoadBox");
		GlobalData.bStayChildWindow = true;
	} else {
		ret = LocalInit();
	}

	RegisterBangs();

	AtVarExec(GlobalData.OnInit,false,false);
	
	TRACE(TRACE_STATUS,"done with initModuleEx");
	return ret;
}

struct CleanupWindow
{
  HWND parent;
  s_WinData *pParent;

  bool operator()(s_WinData *pData)
  {
		if (pData->bHidden)
    {
			parent = GetParent(pData->hWnd);
			pParent = GetWindowPtr(parent);
      if (pParent->hWnd == parent)
      {
			  /* this will probably happen multiple times */ 
			  /* for each parent, but that's ok           */ 
			  MoveWindowToDesk(pParent,GlobalData.CurrentDesk,NULL);
			  MoveWindowToDesk(pData,GlobalData.CurrentDesk,NULL);
			  /* make sure they get shown                 */ 
			  ShowWindow(pData->hWnd,SW_SHOWNA);
			  pData->bHidden = false;
      }
		}
		if (pData->hSnapShot)
    {
			LASTERROR(DeleteObject(pData->hSnapShot));
			StatGDIDel(Bitmap);
			pData->hSnapShot = NULL; /* sorta used as flag */ 
		}
    return false;
  }
};

/*                                                  --------- quitModule
	Input:
		dllInst = my instance
	Output:
		return 0 on success
*/
int quitModule(HINSTANCE dllInst)
{
	int ret=0;

#ifdef VWM_STATS
	ASSERTV(VWMStats.WndProcLevel==0);
	if (VWMStats.WndProcLevel!=0) {
		char buf[256];
		sprintf(buf,"WndProc at %10d never returned.",VWMStats.WndProcStart-VWMStats.InitStart);
		TRACE(TRACE_ERROR,buf);
	}
#endif

	AtVarExec(GlobalData.OnQuit,false,false);

	TRACE(TRACE_STATUS,"Cleaning up for any *Fix windows.");
	/* gather all *Fix windows and their parents          */ 
	/* maybe I should just remember them with LM_SAVEDATA */ 
	/* if VWMHideTaskOnSwitch is set, show all windows hidden by that first */ 
	if (GlobalData.bHideTaskOnSwitch) {
		GlobalData.bHideTaskOnSwitch = false;
		GlobalData.bUnHideTaskOnSwitch = true;
		SwitchDesk(GlobalData.CurrentDesk);
		GlobalData.bUnHideTaskOnSwitch = false;
	}
  CleanupWindow cw;
  IterateAnyOrder(cw);

	if(!GlobalData.bReturnToFirstScreen) {
		SendMessage(GlobalData.hLiteStepWnd, LM_SAVEDATA,
			(WPARAM) (sizeof(int) | VWM_DESKNO), /* length and ident */ 
			(LPARAM) &GlobalData.CurrentDesk);   /* data             */ 
	} else {
		SwitchDesk(0);
	}
	if (!GlobalData.bNoGathering) Gather(GlobalData.CurrentDesk);

	if (GlobalData.hMainWnd!=NULL) ret = LocalQuit();
	ASSERTV(GlobalData.hMainWnd==NULL);

#if defined(RABIDVWM_MULTIMON) && defined(NO_RCMM)
	LSQuitMultiMon();
#endif

	RemoveBangs();

#ifdef USE_LS246_LCOPEN_BUG
	if (GlobalData.ConfigFile) {
		TRACE(TRACE_STATUS,"Freeing config filename.");
		delete[] (void *)GlobalData.ConfigFile;
		DCLEAR(GlobalData.ConfigFile);
	}
#endif

  /* this should allow drag&drop to work while threaded */ 
  CoUninitialize();

	TRACE(TRACE_STATUS,"Unregisterring window class.");
	ASSERTV(dllInst == GlobalData.hInstance);
	UnregisterClass(szAppName,GlobalData.hInstance);

	if (GlobalData.KnownWindows) {
		TRACE(TRACE_STATUS,"Freeing KnownWindows.");
		delete[] GlobalData.KnownWindows;
		DCLEAR(GlobalData.KnownWindows);
	}
#ifndef NDEBUG
	GlobalData.nKnownWindows = 0;
#endif

	rUnlinkFunctions(DynFuncs);

#if defined(VWM_STATS) && defined(TRACE_FILE)
	DumpStats();
#endif

#if !defined(NDEBUG) && defined(TRACE_FILE)
	TraceStartTime = 0;
	TRACE(TRACE_STARTUP,"Bye.");
#endif

	return ret;
}

/*                                                  --------- initWharfModule
*/
int initWharfModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType *wd)
{
	return initModuleEx (ParentWnd, dllInst, (wd) ? wd->lsPath : NULL);
}

/*                                                  --------- quitWharfModule
*/
int quitWharfModule(HINSTANCE dllInst)
{
	return quitModule(dllInst);
}


// *************************************************************************************
// end of litestep exports


void SetOnTop(void)
{
  /* Remove WS_POPUP and make it a WS_CHILD */ 
	SetWindowLong(GlobalData.hMainWnd, GWL_STYLE,
		(GetWindowLong(GlobalData.hMainWnd, GWL_STYLE) &~ WS_POPUP)|WS_CHILD);

  SetVWMPos();

	if (GlobalData.bStayChildWindow) {
		/* in box, set parent to box window */ 
		SetParent(GlobalData.hMainWnd, GlobalData.hParentWnd);
		/* and leave as WS_CHILD            */ 
	} else {
		/* Set the Parent window to Desktop for "onBottom" emulation. */ 
		/* Or set the Parent to NULL if "onTop" is desired            */ 
		SetParent(GlobalData.hMainWnd, GlobalData.bAlwaysOnTop ? NULL : GlobalData.hParentWnd);
		/* Now Set the Window Style back to WS_POPUP                  */ 
		/* so things function correctly.                              */ 
		SetWindowLong(GlobalData.hMainWnd, GWL_STYLE,
		 (GetWindowLong(GlobalData.hMainWnd, GWL_STYLE) &~ WS_CHILD)|WS_POPUP);
	}
}

// refresh settings, don't lose windows
void Refresh(void)
{
	if (GlobalData.hMainWnd!=NULL) LocalQuit();
	/* if loaded as a child, don't reinit if the parent window doesn't exist */ 
  /* err... there was a big problem with refreshing in a box, but i don't  */ 
  /* remember if this fixes it...  Oh, how I wish I'd commented that.      */ 
	if (!GlobalData.bStayChildWindow ||
		(GlobalData.hParentWnd!=NULL && IsWindow(GlobalData.hParentWnd)))
		LocalInit();
	TRACE(TRACE_STATUS,"Refresh complete.");
}

struct ClearWindowFlash
{
  bool operator()(s_WinData *pData)
  { pData->bFlash = false; return false; }
};

int LocalInit(void)
{
	const static UINT Msgs[] = {
		LM_GETREVID,
		LM_BRINGTOFRONT,
		LM_SWITCHTON,
		LM_REFRESH,
		LM_WINDOWCREATED,
		LM_WINDOWDESTROYED,
		LM_WINDOWACTIVATED,
		LM_REDRAW, /* to get window flashes */ 
		0
	};

#ifndef NDEBUG
	bInLocalInit = true;
#endif

	TRACE(TRACE_STATUS,"->LocalInit();");

	if (GlobalData.hMainWnd!=NULL)
	{
		TRACE(TRACE_ERROR,"Main window already exists.");
		LocalQuit();
	}

	if (GetSystemMetrics(SM_CMONITORS)>1) {
		GlobalData.ScreenSize.cx = GetSystemMetrics(SM_CXVIRTUALSCREEN);
		GlobalData.ScreenSize.cy = GetSystemMetrics(SM_CYVIRTUALSCREEN);
		GlobalData.ScreenOffset.x = -GetSystemMetrics(SM_XVIRTUALSCREEN);
		GlobalData.ScreenOffset.y = -GetSystemMetrics(SM_YVIRTUALSCREEN);
	} else {
		GlobalData.ScreenSize.cx = GetSystemMetrics(SM_CXSCREEN);
		GlobalData.ScreenSize.cy = GetSystemMetrics(SM_CYSCREEN);
		GlobalData.ScreenOffset.x = GlobalData.ScreenOffset.y = 0;
	}

	TRACE(TRACE_STATUS,"Reading config.");
	ReadConfig();

	SetDeskCount(GlobalData.nDesksX,GlobalData.nDesksY);

	TRACE(TRACE_STATUS,"First UpdateWindowList();");
	GlobalData.bScaleChanged = true;
	GlobalData.nSetupPass = 1;
	UpdateWindowList();
	/* all windows will flash at first, turn this off               */ 
	/* (doesn't need to be done if we're Refresh()ing, but oh well) */ 
  ClearWindowFlash cwf;
  IterateAnyOrder(cwf);

	if (GlobalData.bGuessDeskCount || GlobalData.bGuessDeskNo) {
		POINT n,x;
		GuessDesksFromWindows(&n,&x);
		/* this should almost never happen */ 
		if (GlobalData.bGuessDeskCount) {
			TRACE(TRACE_WARNING,"Using guessed desk count.");
			int i,j;
			i = x.x-n.x+1; j = x.y-n.y+1;
			/* if we're guessing and it ends up  */ 
			/* being only 1 desk, do 2x2 instead */ 
			if (i<2 && j<2) i=j=2;
			SetDeskCount(i,j);
		}
		/* if not after a recycle, and there are windows outside of */ 
		/* the current desk, assume we're starting after a litestep */ 
		/* crash and use the guessed current desk                   */ 
		if (GlobalData.bGuessDeskNo) {
			if ((x.x-n.x)!=0 || (x.y-n.y)!=0) {
				TRACE(TRACE_WARNING,"Using guessed currentdesk.");
				/* clamp to range to prevent SwitchDesk from being naughty */ 
				if ((-n.x)>=GlobalData.nDesksX) n.x=-(GlobalData.nDesksX-1);
				if ((-n.y)>=GlobalData.nDesksY) n.y=-(GlobalData.nDesksY-1);
				GlobalData.CurrentDesk = (-n.y)*GlobalData.nDesksX+(-n.x);
				UpdateWindowList();
			}
		}
		GlobalData.bGuessDeskCount =
		GlobalData.bGuessDeskNo = false;
	}
	/* now scan again to do *window matches */ 
	GlobalData.nSetupPass = 2;
  {
    SIZE zero = {0,0};
    MoveAllWindows(&zero,GlobalData.CurrentDesk);
  }
	UpdateWindowList();

  if (GlobalData.InitialDesk!=DESK_NONE)
    SwitchDesk(GlobalData.InitialDesk);
  else
  	AtVarExec(GlobalData.OnSwitchTo,false,false);

	GlobalData.nSetupPass = 0;

	/* Used to grab background here for .snapshot and   */ 
	/* FakeTrans, now it's done in WM_CREATE via Resize */ 

	TRACE(TRACE_STATUS,"Creating main window");
	if (GlobalData.bStayChildWindow) GlobalData.bAlwaysOnTop = false;

  /* Always created as if VWMAlwaysOnTop is set */ 
  /* we apply it later anyway, and this avoids  */ 
  /* crashes with VWMTransparent                */ 

	/* GlobalData.hMainWnd = Set on WM_CREATE. */ 
	CreateWindowEx(
		WS_EX_TOOLWINDOW|WS_EX_TOPMOST|(GlobalData.bTransparent?WS_EX_LAYERED:0),
		szAppName,
		NULL,
		WS_POPUP,
		GlobalData.VWMPos.x,   GlobalData.VWMPos.y,
		GlobalData.VWMSize.cx, GlobalData.VWMSize.cy,
		NULL,
		NULL,
		GlobalData.hInstance,
		NULL);

  if (!GlobalData.hMainWnd) 
	{
		LASTERROR(0/*CreateWindowEx*/);
		MessageBox(GlobalData.hParentWnd,
			"Error creating window.",
			szAppName, MB_OK|MB_ICONERROR);
		PostQuitMessage(0);
#ifndef NDEBUG
		bInLocalInit = false;
#endif
		return 1;
	}

  SetVWMPos();

	/* set the magic value to make this window recognized as a part of LiteStep */ 
	TRACE(TRACE_STATUS,"SetWindowLong");
	SetWindowLong(GlobalData.hMainWnd,GWL_USERDATA,magicDWord);

	/* set parent/popup state properly for always on top                        */ 
	TRACE(TRACE_STATUS,"SetOnTop");
	SetOnTop();

	if (!GlobalData.bNoShow) {
		TRACE(TRACE_STATUS,"ShowWindow");
		ShowWindow(GlobalData.hMainWnd,SW_SHOWNA);
	}
	
	TRACE(TRACE_STATUS,"Registering with LiteStep.");
	SendMessage(GlobalData.hLiteStepWnd, LM_REGISTERMESSAGE,
		(WPARAM) GlobalData.hMainWnd, (LPARAM) Msgs);

#ifdef VWMDROPTARGET
	GlobalData.pDropTarget = new VWMDropTarget;
	GlobalData.pDropTarget->Register(GlobalData.hMainWnd);
#else
	TRACE(TRACE_STATUS,"DragAcceptFiles");
	DragAcceptFiles(GlobalData.hMainWnd, TRUE);
#endif

	TRACE(TRACE_STATUS,"Setting refresh timer.");
	LASTERROR(
	GlobalData.RefreshTimerID =
		SetTimer(GlobalData.hMainWnd, TIMER_REFRESH, GlobalData.RefreshTime, NULL)
	);

#ifndef NDEBUG
	bInLocalInit = false;
#endif
#ifdef VWM_STATS
	VWMStats.InitTime = GetTickCount() - VWMStats.InitStart;
#endif
	return 0;
}

int LocalQuit(void)
{
	TRACE(TRACE_STATUS,"->LocalQuit()");

	AtVarExec(GlobalData.OnSwitchFrom,false,false);

	if (GlobalData.hMainWnd) {
		TRACE(TRACE_STATUS,"Killing refresh timer.");
		if (GlobalData.RefreshTimerID) {
			LASTERROR(KillTimer(GlobalData.hMainWnd, GlobalData.RefreshTimerID));
			GlobalData.RefreshTimerID = 0;
		}
#ifdef VWMDROPTARGET
		TRACE(TRACE_STATUS,"Revoking and deleting VWMDropTarget");
		GlobalData.pDropTarget->Revoke(GlobalData.hMainWnd);
		delete GlobalData.pDropTarget;
#endif
		/* if inbox, we're already in WM_DESTROY */ 
		if (!GlobalData.bLoadInBox)
		{
			TRACE(TRACE_STATUS,"Destroying main window.");
			LASTERROR(DestroyWindow(GlobalData.hMainWnd));
		}
	}
	GlobalData.hMainWnd = NULL; /* used as flag */ 

	TRACE(TRACE_STATUS,"Deleting back buffer.");
	if (GlobalData.bmBuffer) {
		LASTERROR(DeleteObject(GlobalData.bmBuffer));
		DCLEAR(GlobalData.bmBuffer);
		StatGDIDel(Bitmap);
	}
	if (GlobalData.bmBackSnapShot) {
		TRACE(TRACE_STATUS,"Deleting Back SnapShot");
		LASTERROR(DeleteObject(GlobalData.bmBackSnapShot));
		DCLEAR(GlobalData.bmBackSnapShot);
		StatGDIDel(Bitmap);
	}
	if (GlobalData.bmSelSnapShot) {
		TRACE(TRACE_STATUS,"Deleting Sel SnapShot");
		LASTERROR(DeleteObject(GlobalData.bmSelSnapShot));
		DCLEAR(GlobalData.bmSelSnapShot);
		StatGDIDel(Bitmap);
	}

	TRACE(TRACE_STATUS,"Deleting PaintDesktopEx cache.");
	FreeDesktopBmp();

	TRACE(TRACE_STATUS,"Releasing config.");
	ReleaseConfig();

	return 0;
}

#ifndef NDEBUG
void AssertFail(const char *assert,const char *file,int line)
{
	char buffer[MAX_LINE_LENGTH];

#ifdef TRACE_FILE
	sprintf(buffer,"Assert failed\nat line %d of %s:\nASSERT(%s);\n",
		line,file,assert);

	TRACE(TRACE_ASSERT,buffer);
#else
	sprintf(buffer,"Assert failed at line %d of %s:\n\nASSERT(%s);\n\nDebug?",
		line,file,assert);

	if (MessageBox(GlobalData.hMainWnd,buffer,szAppName,MB_YESNO)==IDYES)
		/* breakpoint */ 
		DebugBreak();
#endif
}
#endif

#if !defined(NDEBUG) && defined(TRACE_FILE)

#include <stdio.h>

void Trace(const char *level,const char *line)
{
	static bool log=true;
	if (!log) return;
#ifdef VWM_STATS
	DWORD start = GetTickCount();
#endif
	FILE *f;
	f = fopen(TRACE_FILE,"at");
	if (!f) {
		MessageBox(NULL,
			"Couldn't open log file.\nLogging will be turned off.",
			szAppName,MB_OK|MB_ICONERROR);
		log = false;
	}
	if (level==TRACE_STARTUP && lstrcmp(line,"Hello.")==0)
	{
		fprintf(f,"<link rel=\"StyleSheet\" href=\"rabidvwmlog.css\" type=\"text/css\"><hr>");
	}
	fprintf(f,"<div class=\"%s\">",level);
	fprintf(f,"%10d %s</div>\n",GetTickCount()-TraceStartTime,line);
	fclose(f);
#ifdef VWM_STATS
	VWMStats.TraceTime += GetTickCount()-start;
#endif
}

void TraceLastError(const char *code,const char *file,int line)
{
	char buffer[256];
	DWORD error = GetLastError();
	sprintf(buffer,"System error at line %d of %s: %08X",line,file,error);
	Trace(TRACE_WARNING,buffer);
	Trace(TRACE_WARNING,code);
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,NULL,error,NULL,buffer,256,NULL);
	Trace(TRACE_WARNING,buffer);
}

#endif


// but I like being able to do 30%-20...
int localParseCoordinate(LPSTR strVal, int defaultVal, int maxVal)
{
  LPSTR ptr = strVal;
  char sign = 0;
  int value = 0;
 
  if (!*ptr) return defaultVal;
 
  // skip sign, we'll check it later
  if (*ptr=='-' || *ptr=='+') sign = *ptr++;
 
  // could use atoi, but we need a pointer to the end anyway
  // well, just after the number, so not necessarily the very end
  while (*ptr>='0' && *ptr<='9') 
    value = value*10 + *ptr++-'0';
 
  if (sign=='-') value = -value;
 
  switch (*ptr)
  {
  case 0:
    break;
  case '%': // percent based
    value = value*maxVal/100;
    break;
  case 'c':
  case 'C': // center based
    value += maxVal/2;
    sign = 0;
    break;
  case 'r':
  case 'R':
    value += defaultVal;
    sign = 0;
    break;
  default:
    // something we don't like
    break;
  }
 
  if (sign=='-') value += maxVal;
 
  ++ptr;
  // check for %+/-something
  if (*ptr=='+' || *ptr=='-')
    value += atoi(ptr);
 
  return value;
}
