/*
  RabidVWM LiteStep Module - version 1.0
	"At" Variables: ^@...^

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"

#define SEPCHAR '^'
#define ESCAPE_CHAR '\\'

// copy the contents of (parentheses)
int ParenCopy(char *dest,const char *source)
{
	const char *p1,*p2;
	char *p3;
	int c=0;

	ASSERTV(dest!=NULL);
	ASSERTV(source!=NULL);

	p1 = source+1;
	p2 = strchr(p1,')');
	p3 = dest;

	if (p2) {
		/* copy out the windowspec */ 
		while (p1<p2) { *p3++ = *p1++; c++; }
		/* terminate it            */ 
		*p3 = 0;

		return c;
	}
	return 0;
}

void EscapeAtVar(char *dest,size_t buffer_size,const char *escstring)
{
	char buffer[MAX_LINE_LENGTH];
	char escaped[256];
	const char *sptr;
	char *dptr;
	int space;

	ASSERTV(buffer_size==MAX_LINE_LENGTH);

	/* get characters that need to be escaped     */ 
	sptr = escstring;
	dptr = escaped;
	if (sptr)
	{
		/* always escape the escape char            */ 
		*dptr++ = ESCAPE_CHAR;
		while (*sptr && *sptr!=SEPCHAR)
		{
			/* can't list the SEPCHAR, so specify it  */ 
			/* with the ESCAPE_CHAR, which is escaped */ 
			/* already no matter what                 */ 
			if (*sptr==ESCAPE_CHAR) *dptr++ = SEPCHAR;
			else *dptr++ = *sptr;
			++sptr;
		}
	}
	*dptr = 0;

	/* now process the actual string              */ 
	sptr = dest;
	dptr = buffer;
	space = MAX_LINE_LENGTH-1;

	while (*sptr)
	{
		if (*sptr=='$' && GlobalData.EscapeDollar)
		{
			*dptr++ = GlobalData.EscapeDollar;
			if (!--space) break;
		}
		else if (strchr(escaped,*sptr))
		{
			*dptr++ = ESCAPE_CHAR;
			if (!--space) break;
		}
		*dptr++ = *sptr++;
		if (!--space) break;
	}
	*dptr = 0;
	lstrcpy(dest,buffer);
}

// paste atvar at dest and return ptr to end
// if there are enough, maybe move to a table form
char *ReplaceAtVar(char *dest, const char *atvar, int maxchars, s_Action *pAction)
{
	char buffer[MAX_LINE_LENGTH];
	char *ptr;
	s_WinData *pData = NULL;
	int i;
	POINT mPos;
	char c;
	bool left,right;

	*buffer = 0;

	ASSERTV(dest!=NULL);
	ASSERTV(atvar!=NULL);
	ASSERTV(pAction!=NULL);

	c = tolower(*atvar);
	switch (c)
	{
		case '_': /* yay! */ 
			switch (tolower(*++atvar)) {
			case 'n': lstrcpy(buffer,szAppName); break;
			case 'v': lstrcpy(buffer,szAppVer); break;
			case 'a': lstrcpy(buffer,szAppAuth); break;
			}
			break;
		case '#': /* special character */ 
			switch (tolower(*++atvar)) {
			case 'a':
			case 'A': lstrcpy(buffer,"'"); break;
			case 'c':
			case 'C': lstrcpy(buffer,"^"); break;
			case 'd':
			case 'D': lstrcpy(buffer,"$"); break;
			case 'q':
			case 'Q': lstrcpy(buffer,"\""); break;
      case 'n':
      case 'N': lstrcpy(buffer,"\n"); break;
			}
			break;
		case 'd': /* desk */ 
			switch (tolower(*++atvar)) {
			case 'c': /* current */ 
				/* remember, it's 1 based in the real world */ 
				sprintf(buffer,"%d",GlobalData.CurrentDesk+1);
				break;
			case 'e': /* first empty desk */ 
				lstrcpy(buffer,"0");
				for (i=0;i<GlobalData.nDesks;i++) {
					if (GlobalData.Desks[i].nWnds==0) {
						sprintf(buffer,"%d",i+1);
						break;
					}
				}
				break;
			case 'x': /* count horizontal */ 
				sprintf(buffer,"%d",GlobalData.nDesksX);
				break;
			case 'y': /* count vertical   */ 
				sprintf(buffer,"%d",GlobalData.nDesksY);
				break;
			}
			break;
		case 'k': /* keys */ 
			switch (tolower(*++atvar)) {
			case 'a': /* alt              */ 
				left = 0 != IsKeyDown(VK_LMENU);
				right = 0 != IsKeyDown(VK_RMENU);
				break;
			case 'c': /* ctrl             */ 
				left = 0 != IsKeyDown(VK_LCONTROL);
				right = 0 != IsKeyDown(VK_RCONTROL);
				break;
			case 'm': /* context menu key */ 
				left = right = 0 != IsKeyDown(VK_APPS);
				break;
			case 's': /* shift            */ 
				left = 0 != IsKeyDown(VK_LSHIFT);
				right = 0 != IsKeyDown(VK_RSHIFT);
				break;
			case 'w': /* win key          */ 
				left = 0 != IsKeyDown(VK_LWIN);
				right = 0 != IsKeyDown(VK_RWIN);
				break;
			}
			i = 3;
			switch (tolower(*++atvar)) {
			case 'l': i=1; break;
			case 'r': i=2; break;
			default: --atvar; /* back up if we fail to read an optional char */ 
			}
			if ((i&1)==0) left = false;
			if ((i&2)!=0) left |= right;
			/* result */ 
			if (left) lstrcpy(buffer,"true");
			else lstrcpy(buffer,"false");
			break;
		case 'm': /* mouse */ 
			switch (tolower(*++atvar)) {
			case 'd': /* desk */ 
				sprintf(buffer,"%d",1+DeskFromVWMPoint(pAction->MousePos.x,pAction->MousePos.y));
				break;
			case 'x':
			case 'y':
				mPos = pAction->MousePos;
				switch (tolower(*++atvar)) {
				case 'v': /* virtual screen coords */ 
					VWMToScreen(mPos.x,mPos.y,&mPos);
					break;
				case 's': /* screen coords         */ 
					ClientToScreen(GlobalData.hMainWnd,&mPos);
					break;
				default:
          break;
				}
				sprintf(buffer,"%d",(atvar[-1]=='x')?mPos.x:mPos.y);
				break;
			}
			break;
		case 'p': /* param                   */ 
			if (isdigit(*++atvar))
				sprintf(buffer,"%d",GlobalData.AtParams[*atvar-'0']);
			break;
		case 'c': /* command specific window */ 
			pData = pAction->pData;
		case 'w': /* window                  */ 
			if (c!='c') {
				if (atvar[2]=='(' && ParenCopy(buffer,&atvar[2])) {
					ExecWindowBySpec(buffer,&GetWindowProc,(DWORD)&pData);
					/* if they ask for a windowspec, never give them something else */ 
				} else {
					/* otherwise use current window, if IT exists                   */ 
					pData = GetWindowPtr(GlobalData.hFocusWnd);
					if (pData->hWnd!=GlobalData.hFocusWnd) pData = NULL;
				}
			}
			switch (tolower(*++atvar)) {
			case 'c': /* window class  */ 
				if (pData!=NULL && pData->hWnd!=NULL)
					GetClassName(pData->hWnd,buffer,MAX_LINE_LENGTH);
				break;
			case 'd': /* window desk   */ 
				sprintf(buffer,"%d",(pData)?1+pData->HomeDesk:0);
				break;
			case 'h': /* window handle */ 
				sprintf(buffer,"%d",pData?pData->hWnd:0);
				break;
      case 'o': /* owner handle */ 
        sprintf(buffer,"%d",pData?GetWindow(pData->hWnd,GW_OWNER):0);
        break;
			case 't': /* window title  */ 
				if (pData!=NULL && pData->hWnd!=NULL)
					GetWindowText(pData->hWnd,buffer,MAX_LINE_LENGTH);
				break;
			case 'x': /* position */ 
			case 'y':
        if (!pData) { lstrcpy(buffer,"0"); break; }
				mPos.x = pData->Mini.left+GlobalData.VWMPos.x;
				mPos.y = pData->Mini.top+GlobalData.VWMPos.y;
				switch (tolower(*++atvar)) {
				case 'v': /* virtual screen coords */ 
					mPos.x = pData->Area.left;
					mPos.y = pData->Area.top;
					break;
				case 's': /* screen coords         */ 
					mPos.x = pData->Area.left%GlobalData.DeskSize.cx;
					mPos.y = pData->Area.top%GlobalData.DeskSize.cy;
					break;
				default:
          break;
				}
				sprintf(buffer,"%d",(atvar[-1]=='x')?mPos.x:mPos.y);
				break;
			case 's': /* size */ 
        if (!pData) { lstrcpy(buffer,"0"); break; }
				switch (tolower(*++atvar)) {
				case 'x':
				case 'y':
					mPos.x = pData->Mini.right-pData->Mini.left;
					mPos.y = pData->Mini.bottom-pData->Mini.top;
					switch (tolower(*++atvar)) {
					case 'v': /* virtual screen coords */ 
					case 's': /* screen coords         */ 
						mPos.x = pData->Area.right-pData->Area.left;
						mPos.y = pData->Area.bottom-pData->Area.top;
						break;
					default:
            break;
					}
					sprintf(buffer,"%d",(atvar[-1]=='x')?mPos.x:mPos.y);
					break;
				}
			}
			break;
    case 'i':
      switch (tolower(*++atvar))
      {
      case 'b':
        sprintf(buffer,"%d",GlobalData.bmBuffer);
        break;
      }
      break;
	}
	/* skip windowspecs to see if there's an escape string */ 
	if (*++atvar=='(') while (*atvar!=0 && *atvar!=')') atvar++;
	else --atvar;
	/* escape things if specified                          */ 
	if (*++atvar==ESCAPE_CHAR)
		EscapeAtVar(buffer,MAX_LINE_LENGTH,++atvar);
	else if (GlobalData.EscapeDollar)
		EscapeAtVar(buffer,MAX_LINE_LENGTH,NULL);
	ptr = buffer;
	while (*ptr && maxchars) { *dest++ = *ptr++; --maxchars; }
	*dest = 0;
	return dest;
}

// replace all @vars
void ReplaceAtVars(char *dest,const char *source,int maxchars,s_Action *pAction)
{
	/* I originally wanted to use $ instead of ^ */ 
	/* but LiteStep provided no easy way to pass */ 
	/* the $s through, so I changed it.          */ 
	char *dollar,*ptr;
	bool bInDollar = false;
	int ch;
	/* find dollar             */ 
	while ((dollar=strchr(source,SEPCHAR))) {
		/* copy to this point      */ 
		ch = min(maxchars,dollar-source);
		lstrcpyn(dest,source,ch+1);
		maxchars-=ch; source+=ch; dest+=ch;
		ASSERTV(source==dollar);
		if (maxchars<2) break;

		if (!bInDollar && source[1]=='@') {
			bInDollar = !bInDollar;
			/* found a $@              */ 
			dollar = strchr(++source,SEPCHAR);
			if (!dollar) {
				*dest++ = SEPCHAR;
				maxchars--;
			} else {
				ptr = ReplaceAtVar(dest,++source,maxchars,pAction);
				maxchars -= ptr-dest; dest = ptr;
				source = dollar+1;
				bInDollar = !bInDollar;
			}
		}
		else
		{
			/* copy the $ and continue */ 
			*dest++ = *source++;
			maxchars--;
		}
	}
	lstrcpyn(dest,source,maxchars);
}

void AtVarExec(s_Action *pAction,bool bGetMousePos,bool bWindowFromPoint)
{
	if (pAction==NULL) return;

	char buffer[MAX_LINE_LENGTH];

	ASSERTV(pAction->Command!=NULL);

	TRACE(TRACE_STATUS,"AtVarExec");
	TRACE(TRACE_STATUS,pAction->Command);

#ifdef VWM_STATS
	pAction->nCalled++;
#endif

	/* I don't think this hold thing actually works */ 
	if (pAction->Hold)
	{
		TRACE(TRACE_STATUS,"Action on hold.");
		pAction->Hold--;
		return;
	}

#ifdef VWM_STATS
	pAction->nActivated++;
	pAction->tActivated = GetTickCount()-VWMStats.InitStart;
#endif

	if (bGetMousePos) {
		/* fill action struct          */ 
		GetCursorPos(&pAction->MousePos);
		ScreenToClient(GlobalData.hMainWnd,&pAction->MousePos);
	}
	if (bWindowFromPoint) {
		/* only fill pData if it needs */ 
		/* one based on mouse pos      */ 
		pAction->pData = WindowFromVWMPoint(&pAction->MousePos);
	}

	ReplaceAtVars(buffer,pAction->Command,MAX_LINE_LENGTH,pAction);
	pAction->pData = NULL;
	TRACE(TRACE_STATUS,buffer);
	LSExecute(NULL, buffer, SW_SHOWNORMAL);
}

