/*
  RabidVWM LiteStep Module - version 1.0
	Snapshot Based Painting Functions

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"

static void BeginPaintWnd_SnapShot(HDC dc,PixelBlock &,void *store);
static void PaintWnd_SnapShot(HDC dc,PixelBlock &,void *store,s_WinData *pData,RECT *rFull,RECT *rWnd);
static void EndPaintWnd_SnapShot(HDC dc,PixelBlock &,void *store);

const s_PaintWnd Paint_SnapShot = {
	&BeginPaintWnd_SnapShot,
	&PaintWnd_SnapShot,
	NULL,
	&EndPaintWnd_SnapShot
};

struct s_PaintWnd_SnapShot {
	HDC hGreatAndMighty;
};

static void GetSnapShot_Window(s_WinData *pData)
{
  ASSERT(READABLE(pData,sizeof(s_WinData)));

	int w = pData->Mini.right-pData->Mini.left;
	int h = pData->Mini.bottom-pData->Mini.top;

	/* "rolled up" winamp was crashing the module      */ 
	/* not sure if this fixes the symptom or the cause */ 
	if (w>0 && h>0) {

		{
			WTRACE(TRACE_STATUS,pData,"Grabbing snapshot");

      /* get the dc for the window, clipped */ 
			HDC wdc = GetDCEx(pData->hWnd,NULL,DCX_WINDOW|DCX_NORESETATTRS|DCX_CLIPSIBLINGS);
			HDC bdc = CreateCompatibleDC(wdc);
			StatGDIAdd(DC);
			HBITMAP hbm = CreateCompatibleBitmap(wdc,w,h);
			StatGDIAdd(Bitmap);
			HBITMAP bOld = (HBITMAP)SelectObject(bdc,hbm);

			CTRACE(TRACE_WARNING,!wdc,"Couldn't get window dc.");
			CTRACE(TRACE_WARNING,!bdc,"Couldn't create bitmap dc.");
			CTRACE(TRACE_WARNING,!hbm,"Couldn't create bitmap.");

			if (wdc && bdc && hbm) {
				if (pData->hSnapShot)
				{
					/* copy old image */ 
					HDC odc = CreateCompatibleDC(wdc);
					StatGDIAdd(DC);
					HBITMAP bOld = (HBITMAP)SelectObject(odc,pData->hSnapShot);
					BitBlt(bdc,0,0,w,h,odc,0,0,SRCCOPY);
					SelectObject(odc,bOld);
					DeleteDC(odc);
					StatGDIDel(DC);
				}
				else
				{
					/* draw fake snapshot                              */ 
					/* doesn't matter, but I figure if it has to fail, */ 
					/* at least fail the way people will expect it to  */ 
          /* can't fall back to colored windows tho,         */ 
          /* that hasn't been set up                         */ 
					Rectangle(bdc, 0, 0, w, h);
				}
				/* StretchBlt is FAR too slow */ 
				FastStretchBlt(
					bdc,w,h,
					wdc,
					pData->Area.right-pData->Area.left,
					pData->Area.bottom-pData->Area.top,
					GlobalData.SnapShotLevel-1,
					&GetPixelStd);
				/* kill the old one */ 
				if (pData->hSnapShot)
				{
					DeleteObject(pData->hSnapShot);
					StatGDIDel(Bitmap);
				}
				pData->hSnapShot = hbm; hbm = NULL;
			}

			SelectObject(bdc,bOld);
			if (hbm)
			{
				DeleteObject(hbm);
				StatGDIDel(Bitmap);
			}
			DeleteDC(bdc);
			StatGDIDel(DC);
			ReleaseDC(pData->hWnd,wdc);
		}
	}
}

void PaintSelect_SnapShot(HDC dc,s_DeskData *pDesk,RECT *r)
{
	ASSERT(READABLE(r,sizeof(RECT)));
	ASSERT(READABLE(pDesk,sizeof(s_DeskData)));
	ASSERTV(dc!=NULL);
	ASSERTV(GlobalData.bmSelSnapShot!=NULL);

	HDC bdc = CreateCompatibleDC(dc);
	StatGDIAdd(DC);
	HBITMAP bOld = (HBITMAP)SelectObject(bdc,GlobalData.bmSelSnapShot);

	BitBlt(dc,r->left,r->top,r->right-r->left,r->bottom-r->top,bdc,0,0,SRCCOPY);

	SelectObject(bdc,bOld);
	DeleteDC(bdc);
	StatGDIDel(DC);
}

void PaintDesk_SnapShot(HDC dc, s_DeskData *pDesk, RECT *r)
{
	ASSERT(READABLE(r,sizeof(RECT)));
	ASSERT(READABLE(pDesk,sizeof(s_DeskData)));
	ASSERTV(dc!=NULL);
	ASSERTV(GlobalData.bmBackSnapShot!=NULL);

	HDC bdc = CreateCompatibleDC(dc);
	StatGDIAdd(DC);
	HBITMAP bOld = (HBITMAP)SelectObject(bdc,GlobalData.bmBackSnapShot);

	BitBlt(dc,r->left,r->top,r->right-r->left,r->bottom-r->top,bdc,0,0,SRCCOPY);

	SelectObject(bdc,bOld);
	DeleteDC(bdc);
	StatGDIDel(DC);
}

void PaintVWM_SnapShot(HDC dc)
{
	// paint the desktop backs
	PaintVWM_Generic(dc,&PaintDesk_SnapShot);
}

void BeginPaintWnd_SnapShot(HDC dc,PixelBlock &pb,void *store)
{
	s_PaintWnd_SnapShot *pwss;

	ASSERTV(store!=NULL);
	ASSERT(PAINTWND_BUFFER_SIZE>=sizeof(s_PaintWnd_SnapShot));

	pwss = (s_PaintWnd_SnapShot *)store;

  ASSERT(WRITEABLE(pwss,sizeof(s_PaintWnd_SnapShot)));

	pwss->hGreatAndMighty = CreateCompatibleDC(dc);
	StatGDIAdd(DC);
}


void PaintWnd_SnapShot(HDC dc,PixelBlock &pb,void *store,s_WinData *pData,RECT *rFull,RECT *rWnd)
{
#ifdef VWM_STATS
	int SnapShotStart = GetTickCount();
#endif
	s_PaintWnd_SnapShot *pwss = (s_PaintWnd_SnapShot *)store;
	HBITMAP hbmCrud;

	ASSERTV(store!=NULL);
	ASSERTV(pwss!=NULL);
	ASSERTV(pData!=NULL);
	ASSERTV(rFull!=NULL);
	ASSERTV(rWnd!=NULL);

	/* if we just switched desks it might not have repainted yet */ 
	if (GlobalData.bDelaySnapshot && !GlobalData.bConstantUpdate)
	{
		WTRACE(TRACE_STATUS,pData,"Waiting 150 msec before taking snapshot");
		/* set a timer */ 
		GlobalData.SnapShotTimerID = SetTimer
			(GlobalData.hMainWnd,TIMER_SNAPSHOT,150,NULL); /* .15 sec */ 
		/* don't grab a new snapshot yet */ 
	}
	else
	/* if it's all in the screen */ 
	if (
		GlobalData.MouseCurrent==0 /* not processing any mouse action */ 
		/* BUG: if you hold down the button after switching with the  */ 
		/*		mouse, this prevents snapshots from being taken.        */ 
		&& pData->HomeDesk == GlobalData.CurrentDesk
		)
	{
		/* get the snapshot */ 
		if (GlobalData.SnapShotLevel) GetSnapShot_Window(pData);
	}

	if (pData->hSnapShot) {
		hbmCrud = (HBITMAP)SelectObject(pwss->hGreatAndMighty, pData->hSnapShot);

		BitBlt(
			dc,
			rFull->left, rFull->top,
			rFull->right-rFull->left, rFull->bottom-rFull->top,
			pwss->hGreatAndMighty,
			0, 0, ((GlobalData.bFlashState && pData->bFlash) || pData->bFlashOn)?NOTSRCCOPY:SRCCOPY);

		SelectObject(pwss->hGreatAndMighty, hbmCrud);

	} else {

		/* I have to draw something, make it ugly                      */ 
		/* would be better to drop to the color painter, but we'd need */ 
    /* to have another memory space set up for the pens and such   */ 
		/* not a big deal, but i'm too lazy right now to do it         */ 
		Rectangle(dc, rFull->left, rFull->top, rFull->right, rFull->bottom);
	}
#ifdef VWM_STATS
	VWMStats.SnapShotTime += GetTickCount()-SnapShotStart;
#endif
}

void EndPaintWnd_SnapShot(HDC dc,PixelBlock &pb,void *store)
{
	s_PaintWnd_SnapShot *pwss = (s_PaintWnd_SnapShot *)store;

	ASSERTV(store!=NULL);
	ASSERTV(pwss!=NULL);

	DeleteDC(pwss->hGreatAndMighty);
	StatGDIDel(DC);
}

