/*
  RabidVWM LiteStep Module - version 1.0
	Implementation for Special * Window stuff

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
		Since DoWindowMatches is called during each window list update,
		these will be called for all the windows during each update,
		so they should be quick.
*/

#include "RabidVWM.h"

static HWND GetH_Parent(s_WinData *pData)
{
	return GetParent(pData->hWnd);
}

void Set_IgnoreTitle(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	TRACE(TRACE_STATUS,"Set_IgnoreTitle");
	pData->bIgnoreTitleChange = true;
}

void Clear_IgnoreTitle(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	pData->bIgnoreTitleChange = false;
}

void Set_Sticky(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	TRACE(TRACE_STATUS,"Set_Sticky");
	pData->bSticky = true;
}

void Clear_Sticky(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	pData->bSticky = false;
}

static HWND WINAPI GetOwner(HWND h) { return GetWindow(h, GW_OWNER); }
typedef HWND WINAPI RelativeWndFunc(HWND);

void Set_StartDesk(s_WinData *pData, StarWindowMatchItem *pMatch)
{
  s_WinData *pParentOrGaurdian;
	int nDesk;

	TRACE(TRACE_STATUS,"Set_StartDesk");

	/* move to desk only if new window              */ 
	if (!pData->bNewWindow) return;

	/* can't do anything if no desk specified!      */ 
	if (pMatch->Args==NULL) return;

	/* get desk                                     */ 
  /* relative to where it starts, or relative to current desk? */ 
  nDesk = GetBangDeskNo(GlobalData.CurrentDesk, pMatch->Args, NULL);

	/* if desk number specified was 0, try to guess */ 
  static RelativeWndFunc * const pTry[] = { GetOwner, GetParent, NULL };
  for (RelativeWndFunc * const *ppTry = pTry; nDesk<0; ++ppTry)
  {
    if (*ppTry == NULL) return;
    HWND hRelative = (*ppTry)(pData->hWnd);
    pParentOrGaurdian = GetWindowPtr(hRelative);
    if (pParentOrGaurdian->hWnd == hRelative)
      nDesk = pParentOrGaurdian->HomeDesk;
  }

	MoveWindowToDesk(pData,nDesk,NULL);
}

void Clear_StartDesk(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	/* nothing to do here */ 
	return;
}

void Set_Fix(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	TRACE(TRACE_STATUS,"Set_Fix");
	pData->bFix = true;
	return;
}

void Clear_Fix(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	if (pData->bHidden && pData->bFix) {
		ShowWindow(pData->hWnd,SW_SHOWNA);
		pData->bHidden = false;
	}
	pData->bFix = false;
	return;
}

StarWindowMatchList StarLists[] = {
  /* Key,                    Item list, Set proc,         Clear proc,         handle adjust proc */ 
	{ "*VWMIgnoreTitleChange",			NULL, &Set_IgnoreTitle,	&Clear_IgnoreTitle,	NULL					},
	{ "*VWMSticky",									NULL, &Set_Sticky,			&Clear_Sticky,			NULL					},
	{ "*VWMStartWindowsOnDesktop",	NULL, &Set_StartDesk,		&Clear_StartDesk,		NULL					},
	{ "*VWMFix",										NULL, &Set_Fix,					&Clear_Fix,					&GetH_Parent	},
	{ NULL,													NULL, NULL,							NULL,								NULL					}
};

