#ifndef RABIDVWM_WINDOWLIST_H_INCLUDED
#define RABIDVWM_WINDOWLIST_H_INCLUDED

#if _MSC_VER > 1000
#pragma once
#endif

/* loop over all windows in paint order, back to front   */ 
/* break and return current pData if f() returns true    */ 
/* returns NULL if f() never returned true               */ 
template <typename func_f>
s_WinData *IteratePaintOrder(func_f &f)
{
#ifndef VWM_NO_LOOPDETECT
  int nLeft = GlobalData.nKnownWindows;
#endif
  s_WinData *pData;
  int i;

  /* search from back to front                      */ 
  for (i = GlobalData.FirstWindow;
       i >= 0;
       i = pData->PaintNext)
  {
    pData = &GlobalData.KnownWindows[i];
    if (f(pData))
      return pData;

#ifndef VWM_NO_LOOPDETECT
    if (--nLeft<0)
    {
      TRACE(TRACE_ERROR,"Found loop in paint order");
#if defined(TRACE_FILE)
      DumpWindowList();
#endif
      break;
    }
#endif
  }

#ifndef VWM_NO_LOOPDETECT
  ASSERT(nLeft == 0);
#endif
  return NULL;
}

/* I think 2/3 of the places I use that really want to   */ 
/* iterate front to back, but I think the chain in that  */ 
/* direction is not guaranteed to remain valid outside   */ 
/* of UpdateWindowList()                                 */ 


/* loop over all windows in an arbitrary order (by hwnd) */ 
/* break and return current pData if f() returns true    */ 
/* returns NULL if f() never returned true               */ 
template <typename func_f>
s_WinData *IterateAnyOrder(func_f &f)
{
  int nLeft = GlobalData.nKnownWindows;
  s_WinData *pData = GlobalData.KnownWindows;
  while (--nLeft>=0)
  {
    if (f(pData))
      return pData;
    ++pData;
  }
  return NULL;
}

int FindKnownWindow(HWND hWnd); /* find the index of a window, does not validate slot */ 
s_WinData *GetWindowPtr(HWND hWnd); /* get a pointer to a window, does not validate */ 

#endif /* RABIDVWM_WINDOWLIST_H_INCLUDED */ 
